/*@A (C) 1992 Allen I. Holub                                                */
#include <string.h>
#include "cur.h"
#include "box.h"
#include "video.h"

/*!
 * \file WINCREAT.c
 * This function is used to create a new Curses Window
 *
 * \ingroup Curses
 */

/*--------------------------------------------------------
 * Window creation functions.
 * Standard Functions:
 *
 *    WINDOW	*newwin( lines, cols, begin_y, begin_x )
 *		creates a window
 *
 * Nonstandard Functions:
 *
 *    save()  Area under all new windows is saved (default)
 *  nosave()  Area under all new windows is not saved
 *
 *   boxed()  		Window is boxed automatically.
 * unboxed()  		Window is not boxed (default)
 * def_ground(f,b)	Set default foreground color to f, and background color
 *			to b.
 *--------------------------------------------------------
 */

void save    (WINDOW *w) {  w->Save = 1;	}
void nosave  (WINDOW *w) {  w->Save = 0;	}
void boxed   (WINDOW *w) {  w->boxed  = 1;	}
void unboxed (WINDOW *w) {  w->boxed  = 0;	}

/*!
 * \fn void def_ground(WINDOW *w,int f,int b)
 * Set the default foreground and background colors
 *
 * \param w window to set the forground and background colors
 * \param f Forground color
 * \param b Background color
 */
void def_ground(WINDOW *w,int f,int b)
{
    w->attrib = (f & 0x7f) | ((b & 0x7f) << 3);
}

/*--------------------------------------------------------*/

/*!
 * \fn WINDOW	*newwin(int lines,int cols,int begin_y,int begin_x,void *IoChan,int attrib,int nBoxed)
 * Create a new window
 *
 * \param lines Vertical size including border
 * \param cols Horizontal Size (including border
 * \param begin_y Y coordinate of the upper left hand corner
 * \param begin_x X coordinate of the upper left hand corner
 * \param IoChan Pointer to input channel (used for getc)
 * \param attrib background/forground colors, etc
 * \param nBoxed 0 = No border, 1 = Bordered
 * \return returns pointer to WINDOW object on success, NULL on fail
 */
WINDOW	*newwin(int lines,int cols,int begin_y,int begin_x,void *IoChan,int attrib,int nBoxed)
//--------------------------------------------------------
//int	cols;	  /* Horizontal size (including border)	  */
//int	lines;	  /* Vertical size  (including border)	  */
//int	begin_y;  /* X coordinate of upper-left corner	  */
//int	begin_x;  /* Y coordinate of upper-left corner	  */
//---------------------------------------------------------
{
    WINDOW	*win;
	int nCols,nRows;

    if( !(win = (WINDOW *) malloc( sizeof(WINDOW) )) )
    {
		return win;
    }
	win->IO = IoChan;
	DisplayAttrib(&nCols,&nRows);

    if( cols > nCols )
    {
	cols    = 80;
	begin_x = 0;
    }
    else if( begin_x + cols > nCols )
	begin_x = 80 - cols;

    if( lines > nRows )
    {
		lines   = nRows;
		begin_y = 0;
    }
    else if( begin_y + lines > nRows )
		begin_x = nRows - cols;
	win->dvRow		= begin_y;
	win->dvCol		= begin_x;
    win->x_org	    = begin_x ;
    win->y_org	    = begin_y ;
    win->x_size	    = cols    ;
    win->y_size	    = lines   ;
    win->row	    = 0 ;
    win->col	    = 0 ;
    win->scroll_ok  = 0 ;
    win->wrap_ok    = 1 ;
    win->boxed      = nBoxed ;
	win->Echo   = 1;		
	win->Crmode = 0;		/* If 1, use buffered input	  */
	win->Nl	   = 1;		/* If 1, map \r to \n on input	  */
    win->hidden     = 0 ;
    win->attrib	    = attrib ;
    win->image      = !win->Save ? NULL : savescr( begin_x, begin_x + (cols  - 1) ,
					      begin_y, begin_y + (lines - 1) );
    werase(win);

    if( win->boxed )				/* Must be done last */
    {
        box( win, VERT, HORIZ );	/* Box it first		   */
		win->x_size -= 2;		/* Then reduce window size */
		win->y_size -= 2;		/* so that the box won't   */
		win->x_org  += 1;		/* be overwritten.	   */
		win->y_org  += 1;
        cmove(win, win->y_org, win->x_org );
    }
    return win;
}

/*!
 * \fn void WinSetTitle(WINDOW *w, char *title)
 * Set the title of the WINDOW
 *
 * \param w pointer to the window to title
 * \param title pointer to the string that contains the title
 */
void WinSetTitle(WINDOW *w, char *title)
{
	int x,y,l,i;

	y = w->y_org - 1;
	l = strlen(title);
	x = w->x_org - 1 + (w->x_size + 2 - l) / 2;
	for(i=0;i<l;++i)
	{
		SCREEN[ y ][ x+i ].letter    = title[i];
		SCREEN[ y][ x+i ].attribute = w->attrib;
	}
}

