//******************************************************************************
// Curses Device Driver...allows for use of FILE streams for sending data
// to a curses window
//
//******************************************************************************

#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <cio.h>
#include "CursesDev.h"

/*!
 * \file CursesDev.c
 * Device Driver for Curses Library
 */

static int WOpen(IOCB *pIOCB) __attribute__((section(".lowtext")));
static int WClose(IOCB *pIOCB) __attribute__((section(".lowtext")));
static int WGetC(IOCB *pIOCB) __attribute__((section(".lowtext")));
static int WRead(IOCB *pIOCB,char *b, int n) __attribute__((section(".lowtext")));
static int WPutC(IOCB *pIOCB,int c) __attribute__((section(".lowtext")));
static int WWrite(IOCB *pIOCB,char *b, int n) __attribute__((section(".lowtext")));
static int WStatus(IOCB *pIOCB,int m) __attribute__((section(".lowtext")));
static int WXio(IOCB *pIOCB,int cmd,void *pP) __attribute__((section(".lowtext")));
static int WInit(void) __attribute__((section(".lowtext")));

static H_JVEC WJump = {
	WOpen,		//0
	WClose,		//1
	WGetC,		//2
	WRead,		//3
	WPutC,		//4
	WWrite,		//5
	WStatus,	//6
	WXio,		//7
	WInit		//8
};

/*!
 * \fn static int WOpen(IOCB *pIOCB)
 * Opens up a channel to a window
 *
 * This function opens up the channel, but
 * does not actually create the window.  The
 * user must also call XIO using the CURSESDEV_XIO_CREATEWINDOW
 * command. <br>
 * The device name will be of the form W:\<Window title\>.  The
 * text string "Window title" will be displayed on the title line
 * of the window.
 *
 * This function is reached by using the CioOpen function.
 *
 * \param pIOCB pointer to the IO Control Block created for this channel
 * \return always returns 0
 */
static int WOpen(IOCB *pIOCB)
{
	char *pC,*pN;

	pIOCB->p = NULL;
	pC = strchr(pIOCB->dev_name,':');
	if(pC)
	{
		++pC;
		pN = malloc(strlen(pC)+1);
		strcpy(pN,pC);
		pIOCB->dev_name = pN;
	}
	else
		pIOCB->dev_name = NULL;
	return 0;
}

/*!
 * \fn static int WClose(IOCB *pIOCB)
 * Destroy a channel to WINDOW
 *
 * \param pIOCB pointer to the IO Control Block to be destroyed
 */

static int WClose(IOCB *pIOCB)
{
	if(pIOCB->dev_name) free(pIOCB->dev_name);
	return 0;
}

/*!
 * \fn int WGetC(IOCB *pIOCB)
 * Destroy a channel to WINDOW
 *
 * \param pIOCB pointer to a IO Control Block
 */
static int WGetC(IOCB *pIOCB)
{
	return 0;
}

/*!
 * \fn int WRead(IOCB *pIOCB,char *b, int n)
 * Destroy a channel to WINDOW
 *
 * \param pIOCB pointer to a IO Control Block
 * \param b character pointer to buffer
 * \param n number of spaces in buffer
 * \return returns 0
 */
static int WRead(IOCB *pIOCB,char *b, int n)
{
	return 0;
}

/*!
 * \fn int WPutC(IOCB *pIOCB,int c)
 * Destroy a channel to WINDOW
 *
 * \param pIOCB pointer to a IO Control Block
 * \param c character to output
 */
static int WPutC(IOCB *pIOCB,int c)
{
	WINDOW *pW = (WINDOW *)pIOCB->p;
	waddch(pW,c);
	return 0;
}

/*!
 * \fn int WWrite(IOCB *pIOCB,char *b, int n)
 * Destroy a channel to WINDOW
 *
 * \param pIOCB pointer to a IO Control Block
 * \param b character pointer to buffer
 * \param n number of spaces in buffer
 * \return returns 0
 */
static int WWrite(IOCB *pIOCB,char *b, int n)
{
	int i;
	WINDOW *pW = (WINDOW *)pIOCB->p;
	for(i=0;i<n;++i)
		waddch(pW,b[i]);
	return 0;
}

/*!
 * \fn int WStatus(IOCB *pIOCB,int m)
 * Destroy a channel to WINDOW
 *
 * \param pIOCB pointer to a IO Control Block
 * \param m flag to determine type of status to return
 * \return returns status
 */
static int WStatus(IOCB *pIOCB,int m)
{
	return 0;
}

/*!
 * \fn int WXio(IOCB *pIOCB,int cmd,void *pP)
 * Destroy a channel to WINDOW
 *
 * \param pIOCB pointer to a IO Control Block
 * \param cmd XIO command to be executed
 * \param pP pointer to extra parameters that are device and command dependent
 * \return depends on command being processed
 */
static int WXio(IOCB *pIOCB,int cmd,void *pP)
{
	struct _Wparams *pWP;
	WINDOW *pW;
	int retval = 0;
	switch(cmd)
	{
		case CURSESDEV_XIO_CREATEWINDOW:
			pWP = (struct _Wparams *)pP;

			pW = newwin(pWP->ysize,pWP->xsize,pWP->ypos,pWP->xpos,NULL,FG(pWP->fg) | BG(pWP->bg),1);
			scrollok(pW,pWP->mode);
			if(pIOCB->dev_name)
				WinSetTitle(pW,pIOCB->dev_name);
			pIOCB->p= (void *)pW;
			break;
	}
	return retval;
}

static int WInit(void)
{
	CioAddHandler("W",&WJump);
	return 0;
}

/*!
 * \fn void CursesDevInit(void)
 * Initialize curses device driver
 */

void CursesDevInit(void)
{
	WInit();
}

/*!
 * \fn void InitWParams(struct _Wparams *pWP, int xpos,int ypos, int xsize,int ysize,int fg, int bg, int mode)
 * Initialize Windows Parameter Block
 *
 * \param pWP pointer to windows parameter block to initialize
 * \param xpos X coordinate for upper left hand corner of window
 * \param ypos Y coordinate for upper left hand corner of window
 * \param xsize Horizontal width of window (including border)
 * \param ysize Vertical height of window (including border)
 * \param fg Forground color (color of characters)
 * \param bg Background color (color of screen)
 * \param mode border=1, no border = 0
 */
void InitWParams(struct _Wparams *pWP,int xpos,int ypos, int xsize,int ysize,int fg, int bg, int mode)
{
	pWP->xpos = xpos;
	pWP->ypos = ypos;
	pWP->xsize = xsize;
	pWP->ysize = ysize;
	pWP->fg = fg;
	pWP->bg = bg;
	pWP->mode = mode;
}
