/*******************************************************************
** Device Driver for UARTS on ATmega1281/2561
**
** Craeted July 10, 2007
** by Jim Patchell
**
** Written for WinAVR
**
********************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "system.h"
#include "rs232.h"
#include "task.h"
#include "cio.h"

static void InitPort(SERIAL *pS,unsigned short nTSize,unsigned short nRSize, unsigned short nPort,long nBaudRate);
static int ROpen(IOCB *pIOCB);
static int RGetC(IOCB *pIOCB);
static int RPutC(IOCB *pIOCB,int c);
static int RWrite(IOCB *pIOCB,char *b,int l);
static int RStatus(IOCB *pIOCB,int m);
static int RXIO(IOCB *pIOCB,int cmd,void *pP);
static unsigned short CalculateBaudRate(long nBR);

static int RSerrors;

static SERIAL *gSerial[RS232_N_PORTS];	//need these for interrupt service routines

static H_JVEC RJump = {
	ROpen,		//0
	0,			//1
	RGetC,		//2
	0,			//3
	RPutC,		//4
	RWrite,		//5
	RStatus,	//6
	RXIO,		//7
	RInit		//8
};


static RREG SAdr[2] = {
	{(unsigned char *)&UDR0,(unsigned char *)&UCSR0A,(unsigned char *)&UCSR0B,(unsigned char *)&UCSR0C,(unsigned short *)&UBRR0},
	{(unsigned char *)&UDR1,(unsigned char *)&UCSR1A,(unsigned char *)&UCSR1B,(unsigned char *)&UCSR1C,(unsigned short *)&UBRR1},
};

static int ROpen(IOCB *pIOCB) __attribute__ ((section(".lowtext")));
static int ROpen(IOCB *pIOCB)
{
	if(pIOCB->devnum < RS232_N_PORTS)
	{
		pIOCB->p = (void *)malloc(sizeof(SERIAL));	//allocate space for serial control block
		InitPort((SERIAL *)pIOCB->p,RS232_DEF_T_BUFF_SIZE,RS232_DEF_R_BUFF_SIZE,pIOCB->devnum,RS232_DEF_BAUDRATE);
		gSerial[pIOCB->devnum] = (SERIAL *)pIOCB->p;
	}
	return 0;
}

static int RGetC(IOCB *pIOCB) __attribute__ ((section(".lowtext")));
static int RGetC(IOCB *pIOCB)
{
	int retval;
	char sr;
	SERIAL *pI = (SERIAL *)pIOCB->p;

	PendSemaphore(pI->Rx.pSem,0);
	sr = Disable();		//enter critical section
	if(pI->Rx.nChar)	//are there any characters to recieve
	{
		retval = (int)pI->Rx.buff[pI->Rx.tail++];	//get character	
		if(pI->Rx.tail == pI->Rx.size) pI->Rx.tail = 0;	//check tail pointer
		pI->Rx.nChar--;		//decrement the number of chars in buffer
	}
	else
		retval = -1;	//return error
	Enable(sr);	//exit critical section
	return retval;
}

static int RPutC(IOCB *pIOCB,int c) __attribute__ ((section(".lowtext")));
static int RPutC(IOCB *pIOCB,int c)
{
	char sr;
	SERIAL *pI = (SERIAL *)pIOCB->p;

	sr = Disable();		//enter critical section
	if(pI->Tx.nChar < pI->Tx.size)	//and room?
	{
		pI->Tx.buff[pI->Tx.head++] = (unsigned char)c;
		if(pI->Tx.head == pI->Tx.size) pI->Tx.head = 0;	//check head pointer
		if(!pI->Tx.nChar)	//if first char
			*pI->regs->pUCSRnB |= BIT(UDRIE0);	//enable tx interrupts
		pI->Tx.nChar++;		//increment number of chars in buffer
	}
	Enable(sr);	//exit critical section
	return 0;
}

static int RWrite(IOCB *pIOCB,char *b,int l) __attribute__ ((section(".lowtext")));
static int RWrite(IOCB *pIOCB,char *b,int l)
{
	int i = 0;
	SERIAL *pI = (SERIAL *)pIOCB->p;

	PendSemaphore(pI->Tx.pSem,0);	//blcok access
	for(i=0;i<l;++i)
		RPutC(pIOCB,b[i]);	//output characters
	PostSemaphore(pI->Tx.pSem,0);	//release lock
	return i;
}

static int RStatus(IOCB *pIOCB,int m) __attribute__ ((section(".lowtext")));
static int RStatus(IOCB *pIOCB,int m)
{
	int rv=0;

	SERIAL *pI = (SERIAL *)pIOCB->p;
	switch(m)
	{
		case RS232_STAT_NREC:
			rv = pI->Rx.nChar;
			break;
		case RS232_STAT_TMT:
			if(pI->Tx.nChar == 0) rv = 1;
			else rv = 0;
			break;
	}
	return rv;	//return number of chars in recieve buffer
}

static int RXIO(IOCB *pIOCB,int cmd,void *pP) __attribute__ ((section(".lowtext")));
static int RXIO(IOCB *pIOCB,int cmd,void *pP)
{
	long *pL;
	SERIAL *pI = (SERIAL *)pIOCB->p;

	int rv = 0;
	switch(cmd)
	{
		case RS232_XIO_SETBAUD:
			pL = (long *)pP;
			*pI->regs->pUBRR = CalculateBaudRate(*pL);
			break;
	}
	return rv;
}

static unsigned short CalculateBaudRate(long nBR)
{
	return (unsigned short)( ((SYSTEM_CLOCKRATE / nBR) / 16L) - 1   );
}

static char *PNames[2] = {
	"RX0",
	"RX1"
};

static char *TNames[2] = {
	"TX0",
	"TX1"
};

static void InitPort(SERIAL *pS,unsigned short nTSize,unsigned short nRSize, unsigned short nPort,long nBaudRate)
{
	//----------------------------------------------------------------
	// initialize serial port
	//
	// parameters:
	//	nTSize.........size of transmit buffer
	//	nRSize.........size of recieve buffer
	//	nPort..........port number to initialize
	//	nBaudRate......baudrate of port
	//----------------------------------------------------------------
	int i;

	pS->regs = &SAdr[nPort];
	pS->Tx.buff = malloc(nTSize);
	for(i=0;i<nTSize;++i)
		pS->Tx.buff[i] = 0;
	pS->Tx.head = 0;
	pS->Tx.tail = 0;
	pS->Tx.nChar = 0;
	pS->Tx.size = nTSize;
	pS->Tx.pSem = NewSemaphore(1,SEMAPHORE_MODE_BLOCKING,TNames[nPort]);
	pS->Rx.buff = malloc(nRSize);
	pS->Rx.head = 0;
	pS->Rx.tail = 0;
	pS->Rx.nChar = 0;
	pS->Rx.size = nRSize;
	for(i=0;i<nRSize;++i)
		pS->Rx.buff[i] = 0;
	pS->Rx.pSem = NewSemaphore(0,SEMAPHORE_MODE_BLOCKING,PNames[nPort]);
	*pS->regs->pUCSRnB = 0x00; 	//disable while setting baud rate
	*pS->regs->pUCSRnA = 0x00;	//Normal UART Mode (bits 1 and 0)
	*pS->regs->pUCSRnC = BIT(UCSZ01) | BIT(UCSZ00);	//8 bits, 1 Stop Bit, No Parity
	*pS->regs->pUBRR = CalculateBaudRate(nBaudRate);
	*pS->regs->pUCSRnB = BIT(RXCIE0) | BIT(RXEN0) | BIT(TXEN0); 	//Enable UART
}

int RInit(void) __attribute__ ((section(".lowtext")));
int RInit(void)
{
	CioAddHandler("COM",&RJump);
	return 0;
}

//******************************************************************************
// Interrupt service routines for Serial Ports
//******************************************************************************



static void RxIsr(SERIAL *pI)
{
	while(*pI->regs->pUCSRnA & BIT(RXC0))	//while chars availiable
	{
		if(*pI->regs->pUCSRnA & (BIT(DOR0) ) )  RSerrors++;
		pI->Rx.buff[pI->Rx.head++] = *pI->regs->pUDR;	//get char from UART
		if(pI->Rx.head == pI->Rx.size) pI->Rx.head = 0;
		pI->Rx.nChar++;	//increment number of chars in buffer
		PostSemaphore(pI->Rx.pSem,0);
	}
}

static void TxIsr(SERIAL *pI)
{
	char loop = 1;

	while((*pI->regs->pUCSRnA & BIT(UDRE0)) && loop)	//while space availiable in TX
	{
		if(pI->Tx.nChar)	//while there are characters
		{
			*pI->regs->pUDR = pI->Tx.buff[pI->Tx.tail++];	//stuff char
			if(pI->Tx.tail == pI->Tx.size) pI->Tx.tail = 0;
			pI->Tx.nChar--;	//decrement number of chars in buffer
		}
		else
		{
			//we have run out of characters...disable interrupt
			*pI->regs->pUCSRnB &= ~(BIT(UDRIE0));	//disable tx interrupts
			loop = 0;	//terminate loop
		}
	}
}

void USART0_RX_vect(void) __attribute__ ((naked));
void USART0_RX_vect(void)
{
	SAVE_IRQ();
	++InterruptCount;
	RxIsr(gSerial[0]);
	ExitInterrupt();
	RESTORE_IRQ();
}

void USART0_UDRE_vect(void) __attribute__ ((naked));
void USART0_UDRE_vect(void)
{
	SAVE_IRQ();
	++InterruptCount;
	TxIsr(gSerial[0]);
	ExitInterrupt();
	RESTORE_IRQ();
}

void USART1_RX_vect(void) __attribute__ ((naked));
void USART1_RX_vect(void)
{
	SAVE_IRQ();
	++InterruptCount;
	RxIsr(gSerial[1]);
	ExitInterrupt();
	RESTORE_IRQ();
}

void USART1_UDRE_vect(void) __attribute__ ((naked));
void USART1_UDRE_vect(void)
{
	SAVE_IRQ();
	++InterruptCount;
	TxIsr(gSerial[1]);
	ExitInterrupt();
	RESTORE_IRQ();
}





