/*******************************************************************
** Device Driver for UART On a Spartan 3 Starter Board
**
** Craeted April 25, 2008
** by Jim Patchell
**
** Written for WinAVR
**
********************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "system.h"
#include "xserial.h"
#include "task.h"
#include "cio.h"

static void InitXPort(xSERIAL *pS,unsigned short nTSize,unsigned short nRSize, unsigned short nPort,long nBaudRate);
static int xROpen(IOCB *pIOCB) __attribute__ ((section(".lowtext")));
static int xRGetC(IOCB *pIOCB) __attribute__ ((section(".lowtext")));
static int xRPutC(IOCB *pIOCB,int c) __attribute__ ((section(".lowtext")));
static int xRWrite(IOCB *pIOCB,char *b,int l) __attribute__ ((section(".lowtext")));
static int xRStatus(IOCB *pIOCB,int m) __attribute__ ((section(".lowtext")));
static int xRXIO(IOCB *pIOCB,int cmd,void *pP) __attribute__ ((section(".lowtext")));
static unsigned short xCalculateBaudRate(long nBR);

static xSERIAL *gXSerial[RS232_N_PORTS];	//need these for interrupt service routines

static H_JVEC RJump = {
	xROpen,		//0
	0,			//1
	xRGetC,		//2
	0,			//3
	xRPutC,		//4
	xRWrite,	//5
	xRStatus,	//6
	xRXIO,		//7
	xRInit		//8
};


static xRREG SAdr = {
	(unsigned char *)0xf050,
	(unsigned char *)0xf055,
	(unsigned char *)0xf056,
	(unsigned char *)0xf057,
	(unsigned char *)0xf051,
	(unsigned char *)0xf052,
	(unsigned short *)0xf053
};

static int xROpen(IOCB *pIOCB)
{
	if(pIOCB->devnum < RS232_N_PORTS)
	{
		pIOCB->p = (void *)malloc(sizeof(xSERIAL));	//allocate space for serial control block
		InitXPort((xSERIAL *)pIOCB->p,XRS232_DEF_T_BUFF_SIZE,XRS232_DEF_R_BUFF_SIZE,pIOCB->devnum,XRS232_DEF_BAUDRATE);
		gXSerial[pIOCB->devnum-1] = (xSERIAL *)pIOCB->p;
	}
	return 0;
}

static int xRGetC(IOCB *pIOCB)
{
	int retval;
	char sr;
	xSERIAL *pI = (xSERIAL *)pIOCB->p;

	PendSemaphore(pI->Rx.pSem,0);
	sr = Disable();		//enter critical section
	if(pI->Rx.nChar)	//are there any characters to recieve
	{
		retval = (int)pI->Rx.buff[pI->Rx.tail++];	//get character	
		if(pI->Rx.tail == pI->Rx.size) pI->Rx.tail = 0;	//check tail pointer
		pI->Rx.nChar--;		//decrement the number of chars in buffer
	}
	else
		retval = -1;	//return error
	Enable(sr);	//exit critical section
	return retval;
}


static int xRPutC(IOCB *pIOCB,int c)
{
	char sr;
	xSERIAL *pI = (xSERIAL *)pIOCB->p;

	sr = Disable();		//enter critical section
	if(pI->Tx.nChar < pI->Tx.size)	//and room?
	{
		pI->Tx.buff[pI->Tx.head++] = (unsigned char)c;
		if(pI->Tx.head == pI->Tx.size) pI->Tx.head = 0;	//check head pointer
		if(!pI->Tx.nChar)	//if first char
			*pI->regs->pIER |= BIT(XRS232_IER_TXMT);	//enable tx interrupts
		pI->Tx.nChar++;		//increment number of chars in buffer
	}
	Enable(sr);	//exit critical section
	return 0;
}

static int xRWrite(IOCB *pIOCB,char *b,int l)
{
	int i = 0;
	xSERIAL *pI = (xSERIAL *)pIOCB->p;

	PendSemaphore(pI->Tx.pSem,0);	//blcok access
	for(i=0;i<l;++i)
		xRPutC(pIOCB,b[i]);	//output characters
	PostSemaphore(pI->Tx.pSem,0);	//release lock
	return i;
}

static int xRStatus(IOCB *pIOCB,int m)
{
	int rv=0;

	xSERIAL *pI = (xSERIAL *)pIOCB->p;
	switch(m)
	{
		case RS232_STAT_NREC:
			rv = pI->Rx.nChar;
			break;
		case RS232_STAT_TMT:
			if(pI->Tx.nChar == 0) rv = 1;
			else rv = 0;
			break;
	}
	return rv;	//return number of chars in recieve buffer
}

static int xRXIO(IOCB *pIOCB,int cmd,void *pP)
{
	long *pL;
	xSERIAL *pI = (xSERIAL *)pIOCB->p;

	int rv = 0;
	switch(cmd)
	{
		case XRS232_XIO_SETBAUD:
			pL = (long *)pP;
			*pI->regs->pUBRR = xCalculateBaudRate(*pL);
			break;
	}
	return rv;
}

static unsigned short xCalculateBaudRate(long nBR)
{
	long br;
	br = nBR * 8192L;
	br /= 390625L;
	return (unsigned short)br;
}

static char *PNames[2] = {
	"RX0",
	"RX1"
};

static char *TNames[2] = {
	"TX0",
	"TX1"
};

static void InitXPort(xSERIAL *pS,unsigned short nTSize,unsigned short nRSize, unsigned short nPort,long nBaudRate)
{
	//----------------------------------------------------------------
	// initialize serial port
	//
	// parameters:
	//	nTSize.........size of transmit buffer
	//	nRSize.........size of recieve buffer
	//	nPort..........port number to initialize
	//	nBaudRate......baudrate of port
	//----------------------------------------------------------------
	pS->regs = &SAdr;
	pS->Tx.buff = malloc(nTSize);
	pS->Tx.head = 0;
	pS->Tx.tail = 0;
	pS->Tx.nChar = 0;
	pS->Tx.size = nTSize;
	pS->Tx.pSem = NewSemaphore(1,SEMAPHORE_MODE_BLOCKING,TNames[nPort]);
	pS->Rx.buff = malloc(nRSize);
	pS->Rx.head = 0;
	pS->Rx.tail = 0;
	pS->Rx.nChar = 0;
	pS->Rx.size = nRSize;
	pS->Rx.pSem = NewSemaphore(0,SEMAPHORE_MODE_BLOCKING,PNames[nPort]);
	*pS->regs->pUBRR = xCalculateBaudRate(nBaudRate);
//	*pS->regs->pUBRR = (unsigned short)(0.02097152f * ((double)(nBaudRate)) );
	*pS->regs->pIER = 1; 	//Enable UART
}

int xRInit(void)
{
	CioAddHandler("XR",&RJump);
	EIFR = BIT(INT6);
	EIMSK |= BIT(INT6);
	EIFR = BIT(INT6);
	return 0;
}

//******************************************************************************
// Interrupt service routines for Serial Ports
//******************************************************************************



static void RxIsr(xSERIAL *pI)
{
	while(!(*pI->regs->pRSR & BIT(XRS232_RSR_FIFO_MT)) )	//while chars availiable
	{
		pI->Rx.buff[pI->Rx.head++] = *pI->regs->pUDR;	//get char from UART
		if(pI->Rx.head == pI->Rx.size) pI->Rx.head = 0;
		pI->Rx.nChar++;	//increment number of chars in buffer
		PostSemaphore(pI->Rx.pSem,0);
	}
}

static void TxIsr(xSERIAL *pI)
{
	char loop = 1;


	while(!(*pI->regs->pTSR & BIT(XRS232_TSR_FIFO_FULL)) && loop)	//while space availiable in TX
	{
		if(pI->Tx.nChar)	//while there are characters
		{
			*pI->regs->pUDR = pI->Tx.buff[pI->Tx.tail++];	//stuff char
			if(pI->Tx.tail == pI->Tx.size) pI->Tx.tail = 0;
			pI->Tx.nChar--;	//decrement number of chars in buffer
		}
		else
		{
			//we have run out of characters...disable interrupt
			*pI->regs->pIER &= ~(BIT(XRS232_IER_TXMT));	//disable tx interrupts
			loop = 0;	//terminate loop
		}
	}
}

static void TxDoneIsr(xSERIAL *pI)
{
}

static void (*irqfunc[3])(xSERIAL *pI) = {
	RxIsr,
	TxIsr,
	TxDoneIsr
};

void INT6_vect(void) __attribute__ ((naked));
void INT6_vect(void)
{
	SAVE_IRQ();
	++InterruptCount;
	(*irqfunc[*(gXSerial[0]->regs->pIVR)])(gXSerial[0]);
	EIFR = BIT(INT6);
	ExitInterrupt();
	RESTORE_IRQ();
}







