/////////////////////////////////////////////////////////////////////////////////////////////
// This file is compiled with WinAVR version 4.1.2
//
// These routines mess with the stack
//
// you must use either of these optimizations:
// -O1
// -O3
// -O2
//
// Whatever you do, do NOT use -O0....it will NOT work
//
// /////////////////////////////////////////////////////////////////////////////////////////

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "task.h"
#include "pq.h"
#include "HeapManager.h"

/*!
 *	\file createsemaphore.c
 *	\ingroup MultitaskingKernel
 */

/*@{*/

ECB *TimeoutList;	/*!< List of semaphores to check for timeouts	*/
static ECB *ECBPool;	/*!< List of deleted semaphores to re-allocate	*/

/*!
 *	\fn static ECB *AllocECB(void)
 *	\brief Allocate memory for a new semaphore
 *
 * This function first will check a pool of semaphore objects and
 * if there are any objects there, pick the first one it finds and
 * reuse that one. <br>
 * Otherwise, use malloc to get a new block of memory
 *
 * \return returns a pointer to a block of memory for use as a semaphore, NULL on fail
 */


static ECB *AllocECB(void)
{
	//--------------------------------------------------------------------------
	// Allocate an event control Block
	//	Check the ECB pool, if it is empty, then allocate a new ECB from heap
	//
	// return value:
	//	pointer to ECB on success
	//	NULL on fail
	//--------------------------------------------------------------------------
	ECB *rv;
	
	char sr = Disable();
	if (ECBPool)	//check the pool
	{
		rv = ECBPool;
		ECBPool = ECBPool->Tnext;
		Enable(sr);
	}
	else
	{
		Enable(sr);
		rv = (ECB *)malloc(sizeof(ECB));	//or allocate from the heap
	}
	return rv;
}

/*!
 *	\fn static void FreeECB(ECB *e)
 *	\brief Free memory that was used by a semaphore
 *
 * This function will return the block of memory used by a semaphore
 * back to a pool of semaphore objects
 *
 * \param e pointer to the semaphore object to free
 * 
 */

static void FreeECB(ECB *e)
{
	//--------------------------------------------------------------------------
	// put ECB block into ECB pool (don't bother to return it to the heap
	//
	// parameter:
	//	e...........pointer to ECB block to free
	//--------------------------------------------------------------------------
	char sr;

	sr = Disable();
	if(e->SemaphoreMode)	//is it in timeout list?
	{
		if(e == TimeoutList)
			TimeoutList = e->Tnext;
		if(e->Tprev)
			e->Tprev->Tnext = e->Tnext;
		if(e->Tnext)
			e->Tnext->Tprev = e->Tprev;
	}
	e->Tnext = ECBPool;
	ECBPool = e;
	Enable(sr);
}

/*!
 *	\fn ECB *NewSemaphore(int InitCount,int Mode,char *name)
 *	\brief Create a new semaphore
 *
 * This function creates a new semaphore, by first allocating
 * memory and then initializing the semaphore.
 * 
 * \param InitCount Initial count for semaphore event count
 * \param Mode 0=No Timeout, 1 = Semaphore times out
 * \param name Name given to semaphore
 * \return Pointer to semaphore on success, NULL on fail
 * 
 */

ECB *NewSemaphore(int InitCount,int Mode,char *name)
{
	ECB *r;

	if((r = AllocECB()) != NULL)
		CreateSemaphore(r,InitCount,Mode,name);
	return r;
}

/*!
 *	\fn void CreateSemaphore(ECB *e,int InitCount,int Mode,char *n)
 *	\brief Initialize a new semaphore
 *
 * This function is used to initialize a semaphore object.
 *
 * \param e pointer to semaphore to initialize
 * \param InitCount Initial count for semaphore event count
 * \param Mode 0=No Timeout, 1 = Semaphore times out
 * \param n Name given to semaphore
 * 
 */

void CreateSemaphore(ECB *e,int InitCount,int Mode,char *n)
{
/*
 *                                                                       
 * Semaphore objects
 *
 * CreateSemaphore                                                        
 * parameters:                                                            
 *	e............pointer to an event control block to initialize          
 *	InitCount....Initial value of semaphore count                         
 *	Mode.........0 = Semaphore doesn't time out, 1 = Semaphore Times out  
 *	n............name of semaphore (DEBUG only)                           
 *                                                                        
 */
	char sr;

	e->EventCount = InitCount;
	e->SemaphoreMode = Mode;
	strcpy(e->name,n);					//copy name in
	e->task_h = 0;						//no tasks waiting
	e->task_t = 0;						//no tasks waiting
	//-------------------------------------------------------
	// if Mode==1, then put this semaphore into timeout list
	//-------------------------------------------------------
	sr = Disable();
	if(e->SemaphoreMode == 1)
	{
		if(TimeoutList)	//is list already started?
		{
			e->Tnext = TimeoutList;	//make current head next in line
			e->Tprev = 0;			//these is no previous entry
			TimeoutList->Tprev = e;	//this will be prev for next entry
			TimeoutList = e;			//this is now the new head of list
		}
		else			//no start list
		{
			TimeoutList = e;
			e->Tnext = 0;
			e->Tprev = 0;
		}
	}
	Enable(sr);	//restore interrupts
}

/*!
 *	\fn void DeleteSemaphore(ECB *e)
 *	\brief Delete a semaphore
 *
 * This function is used to delete a semaphore. <br>
 * This is done by first notifying any thread waiting
 * on that semaphore that it has been deleted.
 *
 * \param e pointer to semaphore to delete
 * 
 */

void DeleteSemaphore(ECB *e)
{
/*
 * Delete Semaphore
 *	This function is used to free a Semaphore.  Rather than return the memory
 * to the heap, it is put into a pool where it can be reused later without
 * having to do another malloc.
 *
 *	parameters:
 *		e..........pointer to ECB block that describes the semaphore
 *
 */
	char sr;
	TCB *t;

	//are there any tasks waiting for this sucker?
	sr = Disable();
	while(e->task_h)		//remove all tasks
	{
		t = e->task_h;		//get first task in
		e->task_h = t->next;       //remove from list
		t->next = (TCB *)0;
		t->status = EVENT_DELETED;			//return this value
		Insert(&ActiveTasks,t);	//put task back onto active list
	}
	FreeECB(e);		//put semaphore back in pool
	Enable(sr);
}

/*@}*/
